/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/main.c
 * @brief This is the main file for the gtk+ user interface.
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 *
 * Basic structure of the code is this:
 * main
 * -> search -> saveas -> download
 * -> insert
 * -> directory -> insert
 * -> pseudonyms (create|delete)
 * -> namespace (insert|update)
 * -> about
 *
 * TODO: 
 * 
 * - Anticipating future off-band sharing of gnunet keys,
 *   add possibility to download files by specifying the URI
 *   manually
 **/

#include "config.h"

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "util/getopt.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/xmalloc.h"
#include "applications/afs/afs.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/search.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/gtkui/about.h"
#include "applications/afs/gtkui/directory.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/namespace.h"
#include "applications/afs/gtkui/pseudonyms.h"
#include "applications/afs/encoding/directory.h"
#include "applications/common/printhelp.h"
#include "util/configuration.h"

/**
 * This semaphore can be used to prevent the main window
 * from killing GTK at an unhealty time. It is used by
 * the un-interruptable but GUI-updating insert thread.
 **/
Semaphore * refuseToDie;

/**
 * Provides access to toggling pulldown menu shadings
 **/
GtkItemFactory * itemFactory = NULL;

static GtkWidget * main_window_input_line = NULL;

/**
 * Shows the download window if some dl has been started 
 **/
static void show_dlwindow(GtkButton * button,
			  gpointer dummy) {
  if(dlWindow)
    gtk_widget_show(dlWindow);
}

/**
 * Selects all search results from the current search page.
 **/
static void select_all(GtkButton * button,
		       gpointer dummy) {
  gint pagenr;
  GtkWidget * page;
  ListModel * model;

  pagenr = gtk_notebook_get_current_page(notebook);
  page = gtk_notebook_get_nth_page(notebook,
                                   pagenr);
  if(!page)
    return;
  model = gtk_object_get_data(GTK_OBJECT(page),
                              "MODEL");
  if(model)
    gtk_clist_select_all(GTK_CLIST(model->search_result_list));
}

/**
 * Unselects all search results from the current search page.
 **/
static void select_none(GtkButton * button,
		         gpointer dummy) {
  gint pagenr;
  GtkWidget * page;
  ListModel * model;

  pagenr = gtk_notebook_get_current_page(notebook);
  page = gtk_notebook_get_nth_page(notebook,
                                   pagenr);
  if(!page)
    return;
  model = gtk_object_get_data(GTK_OBJECT(page),
                              "MODEL");
  if(model)
    gtk_clist_unselect_all(GTK_CLIST(model->search_result_list));
}


/**
 * Remove the active page from the search results notebook.
 * The respective search will be stopped as well
 * (by a callback assigned to the page earlier on).
 **/
static void close_search(GtkButton * button,
			 gpointer dummy) {
  gint page;
 
  page = gtk_notebook_get_current_page(notebook);
  gtk_notebook_remove_page(notebook, page);
  /* Need to refresh the widget --
     This forces the widget to redraw itself. */
  gtk_widget_draw(GTK_WIDGET(notebook), NULL);
  
  page = gtk_notebook_get_current_page(notebook);
  if(page<0) {
    GtkWidget * entry;

    entry = gtk_item_factory_get_widget(itemFactory,
                                        "/File/Select all");
    gtk_widget_set_sensitive(entry, FALSE);
    entry = gtk_item_factory_get_widget(itemFactory,
                                        "/File/Select none");
    gtk_widget_set_sensitive(entry, FALSE);
    entry = gtk_item_factory_get_widget(itemFactory,
  	  	  		        "/File/Abort search");
    gtk_widget_set_sensitive(entry, FALSE);
  }
}

/**
 * This method is called whenever the user clicks the
 * search button of the main window. 
 * 
 * @param widget not used
 * @param notebook the notebook where the opened result window will be put.
 **/
static void search(GtkWidget * widget,
		   GtkNotebook * notebook) {
  gchar * searchString;
  gchar * searchStringStart;
  GtkWidget * tmp;
  int i; 

  searchString 
    = gtk_entry_get_text(GTK_ENTRY(main_window_input_line));
  if (searchString == NULL) {
    guiMessage("searchString == NULL\n");
    return;
  }
  searchString = STRDUP(searchString);

  /* Remove heading spaces (parsing bugs if not) */  
  i = strlen(searchString)-1;
  while ( (searchString[i]==' ') && 
	  (i>=0) ) {
    searchString[i] = '\0';
    i--;
  }
  searchStringStart = searchString;
  while (*searchStringStart == ' ')
    searchStringStart++;
  if (*searchStringStart=='\0') {
    guiMessage("No search key given!\n");
    FREE(searchString);
    return;
  }

  /* add a new page in the notebook with the
     search results. getSearchWindow returns
     the page */
  tmp = getSearchWindow(searchStringStart);
  if (tmp != NULL) 
    addToNotebook(searchStringStart,
		  tmp);

  /* reset search line to empty */
  gtk_entry_set_text(GTK_ENTRY(main_window_input_line), 
		     ""); 
  FREE(searchString);
}

/**
 * Exit the application (called when the main window
 * is closed or the user selects File-Quit).
 **/
static void destroyMain(GtkWidget * widget,
			gpointer data) {
  stopCron();
  if (notebook != NULL) {
    int i = 0;    
    while(gtk_notebook_get_nth_page(notebook, 0) != NULL) {
      LOG(LOG_DEBUG, 
	  "DEBUG: Removing search page %d\n", 
	  i++);
      gtk_notebook_remove_page(notebook, 0);
    }
  }
  startCron();
  gdk_threads_leave(); /* avoid deadlock! */
  SEMAPHORE_DOWN(refuseToDie);
  gdk_threads_enter();
  gtk_main_quit(); /* back to main method! */
}

/**
 * Remove all of the root-nodes of a particular type
 * from the directory database.
 *
 * @param unused GTK handle that is not used
 * @param contexts bitmask of the databases that should be emptied.
 **/ 
static void emptyDirectoryDatabaseInd(GtkWidget * unused,
				      unsigned int contexts) {
  emptyDirectoryDatabase(contexts);
}

/**
 * Method called from menu bar "File-Quit". Wrapper around
 * destroy.
 **/ 
static void destroy_stub(void) {
  destroyMain(NULL, NULL);
}

/**
 * The pulldown menus.
 **/
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",         
    NULL, 
    NULL,
    0,
    "<Branch>" },
  { "/File/_Insert",  
    "<control>I", 
    openSelectFile, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/File/Show downloads",      
    "<control>w", 
    show_dlwindow, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/File/Select _all",      
    "<control>a", 
    select_all, 
    0, 
    NULL },
  { "/File/Select _none",      
    "<control><shift>a", 
    select_none, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/File/_Abort search",      
    "<control>D", 
    close_search, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL,        
    NULL, 
    0, 
    "<Separator>" },
  { "/File/_Quit",     
    "<control>Q", 
    destroy_stub,
    0, 
    NULL },
  { "/_Advanced",
    NULL,
    NULL,
    0,
    "<Branch>" },

  { "/Advanced/_Assemble Directory",
    NULL,
    NULL,
    0,
    "<Branch>" },
  { "/Advanced/Assemble Directory/from _search results",
    NULL,
    openAssembleDirectoryDialog,
    DIR_CONTEXT_SEARCH,
    NULL },
  { "/Advanced/Assemble Directory/from _inserted files",
    NULL,
    openAssembleDirectoryDialog,
    DIR_CONTEXT_INSERT,
    NULL },
  { "/Advanced/Assemble Directory/from local _namespaces",
    NULL,
    openAssembleDirectoryDialog,
    DIR_CONTEXT_INSERT_SB,
    NULL },
  { "/Advanced/Assemble Directory/from file identifiers from downloaded _directories",
    NULL,
    openAssembleDirectoryDialog,
    DIR_CONTEXT_DIRECTORY,
    NULL }, 
  { "/Advanced/Assemble Directory/sepx1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/Advanced/Assemble Directory/from _all known file identifiers",
    NULL,
    openAssembleDirectoryDialog,
    DIR_CONTEXT_ALL,
    NULL },

  { "/Advanced/sep1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },

  { "/Advanced/Manage _Pseudonyms",     
    NULL,  
    NULL, 
    0,
    "<Branch>" },
  { "/Advanced/Manage Pseudonyms/_Create new pseudonym",     
    NULL,  
    &openCreatePseudonymDialog,
    0,
    NULL },
  { "/Advanced/Manage Pseudonyms/_Delete pseudonym",     
    NULL,  
    &openDeletePseudonymDialog,
    0,
    NULL },
  { "/Advanced/_Insert into Namespace",     
    NULL,  
    NULL,
    0,
    "<Branch>" },
  { "/Advanced/Insert into Namespace/Select from _search results",     
    NULL,  
    &openAssembleNamespaceDialog, 
    DIR_CONTEXT_SEARCH,
    NULL },
  { "/Advanced/Insert into Namespace/Select from _inserted files",     
    NULL,  
    &openAssembleNamespaceDialog, 
    DIR_CONTEXT_INSERT,
    NULL },
  { "/Advanced/Insert into Namespace/Select from results from downloaded _directories",     
    NULL,  
    &openAssembleNamespaceDialog, 
    DIR_CONTEXT_INSERT,
    NULL },
  { "/Advanced/Insert into Namespace/Select from results from local _namespaces",     
    NULL,  
    &openAssembleNamespaceDialog, 
    DIR_CONTEXT_INSERT_SB,
    NULL },
  { "/Advanced/Insert into Namespace/sepx2",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/Advanced/Insert into Namespace/Select from _all known file identifiers",     
    NULL,  
    &openAssembleNamespaceDialog, 
    DIR_CONTEXT_ALL,
    NULL },
  /*  { "/Advanced/_Update content in Namespace",     
    NULL,  
    NULL,  
    0,
    NULL }, */
  { "/Advanced/_Search Namespace",     
    "<control>S",  
    &searchNamespace,  
    0,
    NULL }, 

  { "/Advanced/sep2",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },

  { "/Advanced/_Reset File Identifiers",
    NULL,
    NULL,
    0,
    "<Branch>" },
  { "/Advanced/Reset File Identifiers/List of _search results",
    NULL,
    emptyDirectoryDatabaseInd,
    DIR_CONTEXT_SEARCH,
    NULL },
  { "/Advanced/Reset File Identifiers/List of _inserted files",
    NULL,
    emptyDirectoryDatabaseInd,
    DIR_CONTEXT_INSERT,
    NULL },
  { "/Advanced/Reset File Identifiers/List of entries in local _namespaces",
    NULL,
    emptyDirectoryDatabaseInd,
    DIR_CONTEXT_INSERT_SB,
    NULL },
  { "/Advanced/Reset File Identifiers/List of files from downloaded _directories",
    NULL,
    emptyDirectoryDatabaseInd,
    DIR_CONTEXT_DIRECTORY,
    NULL }, 
  { "/Advanced/Reset File Identifiers/sepx3",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/Advanced/Reset File Identifiers/_All known file identifiers",
    NULL,
    emptyDirectoryDatabaseInd,
    DIR_CONTEXT_ALL,
    NULL },


/*
  { "/_Options",     
    NULL,  
    NULL, 
    0,
    "<Branch>" },
  { "/Options/Preferences(not impl)",
    NULL,  
    NULL, 
    0, 
    NULL },
*/
  { "/_Help",     
    NULL,  
    NULL,
    0,
    "<LastBranch>" },
  { "/Help/_About",   
    NULL,
    about,
    0, 
    NULL },
};

/**
 * This creates the main window
 **/
static void makeMainWindow() {
  GtkWidget * window;
  GtkWidget * button;
  GtkWidget * hbox;
  GtkWidget * vbox;
  GtkWidget * menubar;
  GtkWidget * table;
  GtkWidget * label;
  GtkWidget * entry;
  GtkAccelGroup * accel_group;

  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);

  /* create main window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL); 
  gtk_window_set_title(GTK_WINDOW(window), 
		       "GNUnet: gtk+ GUI");
  gtk_widget_set_usize(GTK_WIDGET(window), 
		       780, /* x-size */
		       300); /* y-size */
  vbox = gtk_vbox_new(FALSE, 1);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "delete_event",
		      GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "destroy",
		      GTK_SIGNAL_FUNC(destroyMain),
		     NULL);
  gtk_widget_show(vbox);

  
  /* create pulldown menues */
  accel_group = gtk_accel_group_new ();
  itemFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, 
 	 		             "<main>",
				     accel_group);
  gtk_item_factory_create_items(itemFactory, 
				nmenu_items,
				menu_items, 
				NULL);
  gtk_window_add_accel_group(GTK_WINDOW (window), 
			     accel_group);
  menubar = gtk_item_factory_get_widget(itemFactory, 
					"<main>");
  gtk_box_pack_start(GTK_BOX (vbox), 
		     menubar,
		     FALSE, 
		     TRUE, 
		     0);

  /* set some default options and show */
  entry = gtk_item_factory_get_widget(itemFactory,
  				      "/File/Show downloads");
  gtk_widget_set_sensitive(entry, FALSE);
  entry = gtk_item_factory_get_widget(itemFactory,
  				      "/File/Select all");
  gtk_widget_set_sensitive(entry, FALSE);
  entry = gtk_item_factory_get_widget(itemFactory,
  				      "/File/Select none");
  gtk_widget_set_sensitive(entry, FALSE);
  entry = gtk_item_factory_get_widget(itemFactory,
  				      "/File/Abort search");
  gtk_widget_set_sensitive(entry, FALSE);
 
  gtk_widget_show(menubar);


  /* a table to put the search results notebook to */
  table = gtk_table_new(6, 6, TRUE);
  gtk_box_pack_start(GTK_BOX (vbox),
		     table,
		     TRUE,
		     TRUE, 0);
  gtk_widget_show(table);
  
  /* add "notebook" for search results */
  notebook = GTK_NOTEBOOK(gtk_notebook_new());
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook),
			   GTK_POS_TOP);
  gtk_table_attach_defaults(GTK_TABLE(table), 
			    GTK_WIDGET(notebook),
			    0, 6, 0, 6);
  gtk_widget_show(GTK_WIDGET(notebook));

  /* BEGIN of SEARCH BOX CODE */
  /* At the bottom, put the search box */
  hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (vbox),
		     hbox, 
		     FALSE, 
		     FALSE, 
		     0);
  gtk_widget_show(hbox);
 
  /* search entry label */
  label = gtk_label_new("Keyword(s):");
  gtk_box_pack_start(GTK_BOX(hbox),
		     label,
		     FALSE, 
		     FALSE,
		     0);
  gtk_widget_show(label);

  /* search input line */
  main_window_input_line = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), 
		     main_window_input_line, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_signal_connect(GTK_OBJECT(main_window_input_line),
		     "activate",
                     GTK_SIGNAL_FUNC(search),
                     notebook);
  gtk_widget_show(main_window_input_line);

  /* search button */
  button = gtk_button_new_with_label("Search");
  gtk_signal_connect(GTK_OBJECT (button), 
		     "clicked",
		     GTK_SIGNAL_FUNC(search), notebook);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  /* END of SEARCH BOX CODE */


  /* show the main window */
  gtk_widget_show(window);
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseOptions(int argc, 
			char ** argv) {
  int shouldexit = OK;
  int c;

  while (1) {
    int option_index=0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "debug",   0, 0, 'd' },
      { 0,0,0,0 }
    };
    
    c = GNgetopt_long(argc,
		      argv, 
		      "vhdc:L:H:", 
		      long_options, 
		      &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;    
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-gtk v%s\n",
	     VERSION,
	     AFS_VERSION);
      shouldexit = SYSERR;
      break;
    case 'h': 
      printf("USAGE: gnunet-gtk [OPTIONS]\n");
      printHelpDefault();
      shouldexit = SYSERR;
      break;
    default:
      LOG(LOG_FAILURE,
	  "Unknown option %c. Aborting.\n"
	  "Use --help to get a list of options.\n",
	  c);
      shouldexit = SYSERR;    
    } /* end of parsing commandline */
  }
  if (GNoptind < argc) {
    fprintf(stderr,
	    "Invalid arguments: ");
    while (GNoptind < argc)
      fprintf(stderr,
	      "%s ", 
	      argv[GNoptind++]);
    fprintf(stderr, 
	    "\n");
    shouldexit = SYSERR;
  }
  return shouldexit;
}

/**
 * The main function.
 **/
int main(int argc,
         char * argv[]) {
  char * downloadDir;

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return(0);

  downloadDir = getConfigurationString("AFS",
				       "DOWNLOADDIR");
  if(downloadDir != NULL) {
    mkdirp(downloadDir);
    chdir(downloadDir);
  }
  FREENONNULL(downloadDir);
   
  startCron();
  refuseToDie = SEMAPHORE_NEW(1);

  g_thread_init(NULL);
  /*  gdk_threads_init(); */
  gtk_init(&argc, &argv);

  makeMainWindow();
  gdk_threads_enter();
  gtk_main();
  gdk_threads_leave(); 

  LOG(LOG_DEBUG, "DEBUG: GUI leaving...\n");
  SEMAPHORE_FREE(refuseToDie);

  stopCron();
  doneUtil();  
  return 0;
}

/* end of main.c */
